;;; **********************************************************************
;;; Copyright (c) 1995 Heinrich Taube and Tobias Kunze.  All rights
;;; reserved. Use and copying of this software and preparation of
;;; derivative works based upon this software are permitted and may be
;;; copied as long as no fees or compensation are charged for use, 
;;; copying, or accessing this software and all copies of this software
;;; include this copyright notice.
;;; **********************************************************************

(in-package :stella)

(let ((ccl:*warn-if-redefine* nil))
  (defun update-cm-menu ())
  (defun install-cm-menu ()
    (unless (ccl:find-menu "Common Music")
      (let* ((bar (ccl:menubar))
             (pos (or
                   (position "Windows" bar :key #'ccl:menu-title 
                             :test #'string-equal)
                   (length bar))))
        ;; put midi menu under stella.
        #+MIDI
        (ccl:set-menubar (remove cm::%midi-menu% (ccl:menubar)))
        (ccl:set-menubar 
         (append (loop repeat pos collect (pop bar))
                 (list (make-cm-menu))
                 bar)))))
  (defun deinstall-cm-menu ()
    (let ((menu (ccl:find-menu "Common Music")))
      (when menu
        #+MIDI 
        (progn (ccl:remove-menu-items menu cm::%midi-menu%)
               (ccl:set-menubar
                (substitute cm::%midi-menu% menu (ccl:menubar))))
        #-MIDI 
        (ccl:set-menubar (remove menu (ccl:menubar)))
        )))


  (defun tl:show-documentation (pathname &key title
                                      (stream t) (indent 0))
    (declare (ignore indent stream))
    (make-instance 'ccl:fred-window :filename pathname 
                   :view-size #@(396 153)
                   :window-title 
                   (if title (format nil "~:(~A~)" title) "")))
  ;(defun warn-user (&rest args)
  ;  (ccl:message-dialog (apply #'format nil args)))
  )

(defun make-cm-menu ()
  (let ((delimiter (make-instance 'ccl:menu-item 
                     :menu-item-title "-" :disabled t)))
    (make-instance 'ccl:menu
      :menu-title "Common Music" 
      :update-function #'update-cm-menu
      :help-spec "Stella Commands"
      :menu-items 
      (list (make-instance 'ccl:menu-item 
              :menu-item-title "Top Level"
              :command-key #\T
              :menu-item-action #'(lambda () (browser .top-level.))
              :help-spec "Top Level browser.")
            (make-instance 'ccl:menu-item 
              :menu-item-title "Get Info"
              :command-key #\I
              :menu-item-action #'menubar-info
              :help-spec "See information on selected objects.")
            (make-instance 'ccl:menu-item 
              ; :command-key #\L ; goddam mcl takes everything
              :menu-item-title "Design Output"
              :menu-item-action #'(lambda () (worksheet 'output-worksheet))
              :help-spec "Create or Edit output layout.")
            (make-instance 'ccl:menu-item 
              :menu-item-title "Design Envelope"
              :menu-item-action #'(lambda () (enved))
              :help-spec "Create/Edit envelopes.")
            (make-instance 'ccl:menu-item 
              :menu-item-title "Stella"
              :command-key  #\=  ; why doesnt #\( work ?
              :menu-item-action 
              #'(lambda () 
                  (ccl:eval-enqueue
                   (if (stella-p) '(tl:top-level-quit "") '(stella))))
              :help-spec "Start/Quit Stella.")
            delimiter
            (make-new-object-menu)
            (make-instance 'ccl:menu-item 
              :menu-item-title "Duplicate"
              :command-key #\D
              :menu-item-action #'menubar-duplicate
              :help-spec "Duplicate selected objects.")
            (make-instance 'ccl:menu-item 
              :menu-item-title "Load..."
              :command-key #\Y
              :menu-item-action #'menubar-load
              :help-spec "Load file into Stella.")
            (make-instance 'ccl:menu-item 
              :menu-item-title "Import..."
              :menu-item-action #'menubar-import
              :help-spec "Import file into Stella.")
            (make-instance 'ccl:menu-item 
              :menu-item-title "Archive..."
              :menu-item-action #'menubar-archive
              :help-spec "Archive objects to file.")
            delimiter
            (make-instance 'ccl:menu-item 
              :menu-item-title "Output"
              :command-key #\U
              :menu-item-action #'menubar-send-output
              :help-spec "Run window output to current stream.")
            (make-streams-menu)
            #+MIDI cm::%midi-menu%
            delimiter
            (make-instance 'ccl:menu 
              :menu-title "Help"
              :menu-items
              (list
               (let ((h (make-help-menu 
                         "Capella" 
                         (namestring 
                          (merge-pathnames ":doc:help:capella:text.text"
                                           *common-music-directory*))
                         '("About Capella" "about")
                         "Overview" "Selection")))
                 (ccl:add-menu-items 
                  h
                  delimiter
                  (make-instance 'ccl:menu-item 
                    :menu-item-title "Load Tutorial"
                    :style :italic :menu-item-action #'load-capella-tutorial))
                 h)
               (make-instance 'ccl:menu 
                 :menu-title "Stella"
                 :menu-items
                 (list
                  (let ((coms (loop for c in *commands*
                                    collect
                                    (let ((f (tl::command-help-file c)))
                                      (if f (list (tl::command-name c) f )
                                          (tl::command-name c))))))
                    (apply #'make-help-menu "Commands"
                           (namestring 
                            (merge-pathnames ":doc:help:stella:text.text"
                                             *common-music-directory*))
                           coms))
                  (make-help-menu "Topics"
                                  (namestring 
                                   (merge-pathnames ":doc:help:stella:text.text"
                                                    *common-music-directory*))
                                  "expressions" "referencing")))))
            ;delimiter
            (make-instance 'ccl:menu-item 
              :menu-item-title "Preferences..."
              :menu-item-action #'(lambda () (worksheet 'preferences))
              :help-spec "Open Preferences worksheet.")))))

(defun update-cm-menu (menu)
  (let ((item (or (ccl:find-menu-item menu "Stella")
                  (ccl:find-menu-item menu "Quit Stella"))))
    (ccl:set-menu-item-title item (if (stella-p) "Quit Stella" "Stella"))
    ; (ccl:set-command-key item (if (stella-p) #\) #\())
    )
  (if (container-objects .top-level.) 
    (ccl:menu-item-enable (ccl:find-menu-item menu "Design Output"))
    (ccl:menu-item-disable (ccl:find-menu-item menu "Design Output")))
  (ccl:menu-item-enable (ccl:find-menu-item menu "Archive..."))
  (if (archive-ok? *browser*)
    (ccl:menu-item-enable (ccl:find-menu-item menu "Archive..."))
    (ccl:menu-item-disable (ccl:find-menu-item menu "Archive..."))) 
  (if *browser*
    (progn
      (ccl:menu-item-enable (ccl:find-menu-item menu "Get Info"))
      (if (or (and (listing-browser? *browser*)
                   (any-selection? *browser*))
              (not (systemp (object-view-object *browser*))))
        (ccl:menu-item-enable (ccl:find-menu-item menu "Duplicate"))
        (ccl:menu-item-disable (ccl:find-menu-item menu "Duplicate"))))
    (progn
      (dolist (i '("Get Info" "Duplicate"))
        (ccl:menu-item-disable (ccl:find-menu-item menu i)))))

  ;(if (send-output? *browser*)
  ;  (ccl:menu-item-enable (ccl:find-menu-item menu "Output"))
  ;  (ccl:menu-item-disable (ccl:find-menu-item menu "Output")))

  (if (window-send-output? (ccl::front-window))
    (ccl:menu-item-enable (ccl:find-menu-item menu "Output"))
    (ccl:menu-item-disable (ccl:find-menu-item menu "Output")))

  (update-streams-menu (ccl:find-menu-item menu "Streams"))
  (let ((new (ccl:find-menu-item menu "New")))
    (update-algo-menu (ccl:find-menu-item new "Algorithm") 'algorithm)
    (update-algo-menu (ccl:find-menu-item new "Generator") 'generator)))

(defun make-help-menu (title defaults &rest files)
  (make-instance 'ccl:menu
    :menu-title title
    :menu-items
    (loop for file in files
          collect
          (let (s f)
            (if (consp file)
              (setf s (string-capitalize (first file)) 
                    f (merge-pathnames (second file) defaults))
              (setf s (string-capitalize file)
                    f (merge-pathnames file defaults)))
            (make-instance 'ccl:menu-item
              :menu-item-title s
              :menu-item-action 
              #'(lambda () (show-documentation f :title s)))))))

(defun make-new-object-menu ()
  (make-instance 'ccl:menu
    :menu-title "New"
    :menu-items
    (list
      (make-instance 'ccl:menu-item
        :menu-item-title "Thread"
        :menu-item-action #'(lambda () (menubar-new 'thread)))
      (make-instance 'ccl:menu-item
        :menu-item-title "Merge"
        :menu-item-action #'(lambda () (menubar-new 'merge)))
      (let ((m (make-instance 'ccl:menu :menu-title "Algorithm")))
        (update-algo-menu m 'algorithm))
      (let ((m (make-instance 'ccl:menu :menu-title "Generator")))
        (update-algo-menu m 'generator))
      (make-instance 'ccl:menu-item
        :menu-item-title "Mute"
        :menu-item-action
        #'(lambda () (menubar-new 'algorithm nil))))))

(defun make-streams-menu ()
  (let ((items '()))
    (map-subobjects 
     #'(lambda (s) 
         (dolist (e (slot-value s 'stream-types))
           (let ((c (find-class (first e))))
             (push 
              (make-instance 'ccl:menu-item
                :menu-item-title
                (string-capitalize
                 (substitute #\Space #\- (symbol-name (first e))))
                :menu-item-action
                #'(lambda (&aux s)
                    (setf s (open-event-stream c))
                    (set-output-stream s)
                    (unless (ccl:command-key-p)
                      (browser s 'info-browser))))
              items))))
     .syntaxes.)
    (make-instance 'ccl:menu :menu-title "Streams" 
                   :menu-items 
                   (list (make-instance 'ccl:menu-item
                           :menu-item-title "-" :disabled t)
                         (make-instance 'ccl:menu
                           :menu-item-title "New Stream"
                           :menu-items (nreverse items))))))
                       
(defun update-streams-menu (menu &optional maker (type 'event-stream))
  (let* ((items (ccl:menu-items menu))
         (all (container-objects .io-streams.))
         (beg (or (position-if #'(lambda (x) (typep x 'object-menu-item))
                               items)
                  (if (typep menu 'ccl::pop-up-menu) 2 0))) ; title items in popup
         (save items)
         (not (find-if-not #'(lambda (x) (typep x 'object-menu-item))
                           items :start beg))
         new edit)
    ;; check for renamed or deleted streams.
   (loop for i in (nthcdr beg items)
          until (eq i not)
          for o = (object-menu-item-object i)
          do
          (unless (same-name? o (ccl:menu-item-title i))
            (unless edit (setf items (copy-list items) edit t))
            (setf items (delete i items)))
          (unless (find o all)
            (unless edit (setf items (copy-list items) edit t))
            (setf items (delete i items))))
    ;; collect new streams.
    (map nil 
         #'(lambda (o &aux (n (object-name o)))
             (unless (find n items :key #'ccl:menu-item-title
                           :test #'string-equal
                           :start beg)
               (when (typep o type)
               (setf edit t)
               (push (if maker (funcall maker o)
                         (make-instance 'object-menu-item
                           :object o :menu-item-title n
                           :menu-item-action
                           #'(lambda ()
                               (set-output-stream o)
                               (when (ccl:command-key-p)
                                 (browser o 'info-browser)))))
                     new))))
         all)
    (when edit
      (setf items 
            (nconc (loop until (eq (car items) not) 
                         collect (pop items))
                   new items))
      (apply #'ccl:remove-menu-items menu save)
      (apply #'ccl:add-menu-items menu items))
 
    ;; mark current output if we are not in a popup
    (unless (typep menu 'ccl::pop-up-menu)
      (loop for i in (nthcdr beg items)
            until (eq i not)
            do (ccl:set-menu-item-check-mark
                i (eq (object-menu-item-object i)
                      *output-stream*))))
    menu))

;;;
;;;
;;;

#|
(defmethod menubar-design-output (&optional (b *browser*))
  (declare (ignore b))
  (let ((w (top-output-window)))
    (declare (dynamic-extent w))   ; huh???? why?
    (if w 
      (progn 
        (get-layout-from-top-listing-view)
        (ccl:window-select w))
      (make-output-window))))
|#

(defun menubar-send-output ()
  (window-send-output (ccl::front-window)))

(defun menubar-new (class &optional (other nil otherp))
  (update-top-level :add
                    (if otherp (make-object class :class other)
                        (make-object class))))

(defun menubar-load (&optional file)
  (when (or file
            (setf file (ccl:choose-file-dialog :button-string "Load")))
    (ccl:with-cursor ccl:*watch-cursor*
      (let ((*merge* t)) ; avoid calling update-top-level until end.
        (load file))
      (update-top-level :update))))

(defun menubar-import (&optional file)
  (when (or file
            (setf file (ccl:choose-file-dialog :button-string "Load")))
    (ccl:with-cursor ccl:*watch-cursor*
      (let ((*merge* t)) ; avoid calling update-top-level until end.
        (import-file file))
      (update-top-level :update))))

(defun menubar-archive ()
  (with-focus-browser (browser)
    (let ((object (object-view-object browser))
          (objects '())
          (selections? nil)
          archive file prompt)
      (when (listing-browser? browser)
        (let ((listing (browser-listing browser)))
          (when (setf selections? (any-selection? listing))
            (map-selected-objects
             #'(lambda (o)
                 (when (event? o)
                   (warn-user "Archive selections must be containers.  ~
                               Clear all selections archive ~A, ~
                               or reselect only containers." 
                              (object-namestring object))
                   (return-from menubar-archive))
                 (push o objects))
             listing)
            (setf objects (nreverse objects)))))
      (if selections?
        (setf file "archive.cm" prompt "Archive selections to:")
        (let ((name (object-namestring object)))
          (setf file (format nil "~(~A~).stella" name)
                prompt (format nil "Archive ~A to:" name)
                objects (list object))))
      (setf archive (ccl:choose-new-file-dialog 
                     :prompt prompt :directory file))
      (ccl:with-cursor ccl:*watch-cursor*
        (archive-object objects archive :verbose nil)
        (update-browser browser)))))

(defun menubar-duplicate ()
  ;; can duplicate if a listing view with selections
  ;; or browser object is a nonsystem container.
  (with-focus-browser (browser)
    (if (and (listing-browser? browser)
             (any-selection? browser))
      (listing-new (browser-listing browser) 1 nil)
      (let ((obj (object-view-object browser)))
        (when (and (typep obj 'id-mixin)
                   (not (systemp obj)))
          (let ((new (duplicate-object obj)))
            (update-top-level :add new)))))))

(defun menubar-info ()
  (open-browser 'info-browser))

(defun load-capella-tutorial ()
  (let ((f (merge-pathnames ":doc:tutorials:capella:Capella.Tutorial"
                            *common-music-directory*))
        (b (merge-pathnames ":doc:tutorials:capella:capella.fasl"
                            *common-music-directory*))
        (s (merge-pathnames ":doc:tutorials:capella:capella.cm"
                            *common-music-directory*)))
    (if f
      (ccl:with-cursor ccl:*watch-cursor*
        (if (probe-file b)
          (if (probe-file s)
            (if (> (file-write-date b) (file-write-date s))
              (load b)
              (load s))
            (load b))
          (if (probe-file s) 
            (load s)
            (warn-user "Can't find ~A!" s)))
        (show-documentation (namestring f)))
      (warn-user "Can't find Capella tutorial: ~A" f))))

(defun update-algo-menu (menu type)
  (map-classes 
   #'(lambda (o &aux s n) 
       (setf s (class-name o))
       (unless (or (eq s 'comment) (eq s 'rest)
                   (ccl:find-menu-item 
                    menu (setf n (symbol-name s))))
         (ccl:add-menu-items 
          menu (make-instance 'ccl:menu-item
                 :menu-item-title (string-capitalize n)
                 :menu-item-action
                 #'(lambda () (menubar-new type s))))))
   'element)
  menu)


;;;
;;;
;;;

(eval-when (load eval)
  (install-cm-menu))

;;;
;;; menu-action-mixin sets the action function of all its menu items to the
;;; generic function menu-action, which takes two arguments: the menu and
;;; the selected item.  the default method for menu-action does nothing,
;;; subclasses of menu or menu-item can specialize as appropriate.
;;;

(defclass menu-action-mixin () ())
(defgeneric menu-action (menu item))

(defmethod initialize-instance :after ((menu menu-action-mixin) &rest ignore)
  (declare (ignore ignore))
  (map nil #'(lambda (item)
               (ccl:set-menu-item-action-function 
                item #'(lambda () (menu-action menu item))))
       (ccl:menu-items menu)))

(defmethod menu-action (menu item)
  (declare (ignore menu item))
  nil)

;;;
;;; This needs to get reimplemented in term of menu-action.
;;; object pop-up menu. menu selection invokes object-menu-action on
;;; the menu and object. the default method sets the object slot of
;;; the menu to the selected item's object.
;;;.

(defgeneric object-menu-action (menu object))

(defclass object-menu (ccl:pop-up-menu)
  ((quiet :initform nil :initarg :quiet-item)
   (update :initform nil :initarg :update)
   (action :initform nil :initarg :action)
   (objects :initform nil :initarg :objects
            :accessor object-menu-objects)
   (object :initform nil :initarg :object 
           :accessor object-menu-object))
  (:default-initargs
    :view-font *capella-font*))

(defmethod initialize-instance :after ((menu object-menu) &rest ignore)
  (declare (ignore ignore))
  (flet ((objmenuitem (menu obj)
           (if (typep obj 'ccl:menu-item) obj
               (make-instance 'object-menu-item
                 :object obj
                 :menu-item-action 
                 #'(lambda () (object-menu-action menu obj)))
               )))
    (let ((objects (slot-value menu 'objects))
          (default (slot-value menu 'quiet))
          (items '()))
      (when default
        (push (make-instance 'ccl:menu-item
                :menu-item-title default
                :style :italic
                :menu-item-action 
                #'(lambda () (object-menu-action menu nil)))
              items))
      (loop for first = t then nil
            for o in objects
            do
            (cond ((consp o)
                   (unless first
                     (push (make-instance 'ccl:menu-item
                             :menu-item-title "-" :disabled t)
                           items))
                   (loop for o in o
                         do (push (objmenuitem menu o) items)))
                  (t (push (objmenuitem menu o) items))))
      (apply #'ccl:add-menu-items menu (nreverse items))
      (setf (slot-value menu 'objects)
            (loop for i in (ccl:menu-items menu 'object-menu-item)
                  collect (object-menu-item-object i))))))

(defmethod object-menu-action ((menu object-menu) object)
  (setf (slot-value menu 'object) object)
  (let ((action (slot-value menu 'action))
        (update (slot-value menu 'update)))
    (when action (funcall action menu object))
    (when update (funcall update (ccl:view-window menu)))))


(defclass object-menu-item (ccl:menu-item)
  ((object :initform nil :initarg :object 
           :accessor object-menu-item-object)))

(defmethod initialize-instance :after ((item object-menu-item) &rest args)
  (declare (ignore args))
  (let ((object (slot-value item 'object)))
    (when object
      (ccl:set-menu-item-title
       item (object-namestring object)))))

;;;
;;; template-menu is a pop up menu containing event classes. selecting
;;; a class installs a slot-table in set worksheet for for entering the
;;; appropriate slot value forms.
;;;

(defclass template-menu (object-menu)
  ((templates :initform nil)
   (tablepos :initform 0 :initarg :table-position))
  (:default-initargs
    :view-font *capella-font*
    :view-nick-name ':template-menu))

(defmethod initialize-instance :after ((menu template-menu) &rest args)
  (declare (ignore args))
  (setf (slot-value menu 'templates)
        (cons (list nil)
              (loop for o in (object-menu-objects menu)
                    collect (list o)))))


#|
MIDI >
CSOUND
STREAMS >



STREAMS
   ---
   ------
   ----
   CSound Scorefile
   Midi File
   MIDI Port


Streams
New Stream

|#
