;;; "mix" is available by anonymous ftp from
;;; ftp://notam.uio.no/pub/sgi/ together with other SGI-related
;;; music-software from NoTAM.  Read the WWW-pages on
;;; http://www.notam.uio.no/ for more info about "mix" and other
;;; software developed at NoTAM.
;;;

(in-package :stella)

(defsyntax sgimix
   :pathname (pathname "test.mix")
   :stream-types '((mix-file "mix")))

(defun volquad (vols quads file)
  ;; volume- and  quad-knob for each track
  (loop for track from 1 to 9
        do
        (format file "volume ~A ~A~%"
                track
                (if (listp vols) (pop vols) vols))
        (format file "quadpos ~A ~A~%"
                track
                (if (listp quads) (pop quads) quads))))

(defun trackgraphs (graph file n-tracks val)
  (loop for track from 1 to n-tracks
        do
        (format file "~A ~A~%~
                      ~9,2F ~A~%~
                      ~9,2F ~A~%~
                      end~2%"
                graph track 0.00 val 960.00 val)))

;;; Handling of input-files to mix:
;;; 
;;; This function is a slightly modified version of 'infiles' that
;;; works across OS's --- pathname-functions sabotages paths and
;;; extensions.  I cant find any logical-pathname functions in excl
;;; (ACL 3.1), so here we use merging of strings instead, with
;;; explicit searching for logical units:

;; These next variables may be setf'd for working with input-files for
;; mcl-programs or pc-programs or whatever.

(defvar default-in-file "/sounds/test.aiff")
(defvar path-delimit "/")

;; Merges file-name-strings with default-name.  Takes either a
;; string of names separated by comma (eg. "test4.aiff,
;; andersvi/test2"), or several separated strings, with the same
;; possibility of comma-separating within a single string.  Paths
;; and extensions are sticky all the way through:
;; 
;; e.g.:  (insounds "sinsen/trikken, bussen.aiff" "/banen/die/1, 2.aifc") =>
;; 
;; '("/sounds/sinsen/trikken.aiff" "/sounds/sinsen/bussen.aiff"
;;                     "/banen/die/1.aiff" "/banen/die/2.aifc")

(defun insounds (str &rest others)
  (flet ((merge-names (name default)
	    (let* ((def-path (subseq default 0 (1+ (search path-delimit default :from-end t))))
		   (def-affix (subseq default (search "." default :from-end t)))
		   (name-path (if (search path-delimit name :start2 0 :end2 1) "" def-path))
		   (name-affix (if (search "." name :from-end t) "" def-affix)))
	      ;; merge the resultant pieces with name
	      (concatenate 'string name-path name name-affix)))
	 (elements (namelist)
	    (let ((out '()))
	      (dolist (i namelist)
		      (setf out (append (sequence-to-list i :read nil) out)))
	      out)))
    (let* ((the-others (elements others))
	   (list (append (sequence-to-list str :read nil) the-others))
	   file)
      (setf file (merge-names (pop list) default-in-file))
      (cons file
	    (loop for f in list
		  collect (setf f (merge-names f file))
		  do (setf file f))))))

(defclass mix-file (event-file)
  ((mixsounds :initarg mixsounds :initarg :mixsounds :initform nil)
   (srate :initarg srate :initarg :srate :initform 44100)
   (chns :initarg chns :initarg :chns :initform 2)
   (bufsiz :initarg bufsiz :initarg :bufsiz :initform 0.50)
   (linvol :initarg linvol :initarg :linvol :initform 0)
   (showlin :initarg showlin :initarg :showlin :initform 0)
   (shownum :initarg shownum :initarg :shownum :initform 0)
   (showgrid :initarg showgrid :initarg :showgrid :initform 0)
   (showallgraphs :initarg showallgraphs :initarg :showallgraphs :initform 0)
   (autoscroll :initarg autoscroll :initarg :autoscroll :initform 0)
   (fastdraw :initarg fastdraw :initarg :fastdraw :initform 0)
   ;; mastervolume is global (for all tracks) volume=60 is full burst.
   (mastervolume :initarg mastervolume :initarg :mastervolume :initform 60)
   ;; pr. track data:
   (volumes :initarg volumes :initarg :volumes :initform 60)
   (quads :initarg quads :initarg :quads :initform 1) ;1 or 2 for front and behind.
   (volgraphs :initarg volumegraphs :initarg :volumegraphs :initform nil)
   (pangraphs :initarg pangraphs :initarg :pangraphs :initform nil)
   (auxgraphs :initarg auxgraphs :initarg :auxgraphs :initform nil)
   (syntax :initform (find-syntax ':sgimix))))

(defmethod initialize-stream-for-processing ((stream mix-file))
  (let ((file (slot-value stream 'stream)))
    ;; comments begin and end with 1 single backslash (escaped here).
    (format file "\\ Output from Common Music - ~A \\~%" (cm::date-string))
    (format file "samplerate ~A~%~
                  channels ~A~%~
                  buffersize ~A~%~
                  linearvol ~A~%~
                  showlinear ~A~%~
                  shownumbers ~A~%~
                  showgrid ~A~%~
                  showallgraphs ~A~%~
                  autoscroll ~A~%~
                  fastdraw ~A~2%"
            (slot-value stream 'srate)
            (slot-value stream 'chns)
            (slot-value stream 'bufsiz)
            (slot-value stream 'linvol)
            (slot-value stream 'showlin)
            (slot-value stream 'shownum)
            (slot-value stream 'showgrid)
            (slot-value stream 'showallgraphs)
            (slot-value stream 'autoscroll)
            (slot-value stream 'fastdraw))))

(defmethod deinitialize-stream-for-processing ((stream mix-file))
  (let ((file (slot-value stream 'stream)))
    (format file "~%")
    (volquad (slot-value stream 'volumes) (slot-value stream 'quads) file)
    ;; mastervolume is a one-timer in track 10
    (format file "volume 10 ~A~2%" (slot-value stream 'mastervolume))
    ;; - graphtype, stream, n-tracks, value
    (trackgraphs "volumegraph" file 10 60)
    (trackgraphs "pangraph" file 9 30)
    (trackgraphs "auxgraph" file 10 0)))

(defmethod post-process-stream ((stream mix-file) inits)
  (declare (ignore inits))
  (let ((file (namestring (slot-value stream 'pathname)))
        doit)
    (setf doit
          (if *command-prompting*
              (ask-y-or-n :prompt (format nil "open file ~A in mix? " file)
                          :default 'no)
            nil))
    (when doit (cm::shell (format nil "~Aunix/playsgimix ~A"
				  *common-music-directory* file)))))


;;;
;;; The "sound" object.  One may choose whether to specify input-files
;;; to mix when opening a processing-stream (like in the rt-syntax),
;;; or as data in the sound-object.  The filter in the method on
;;; write-event collects sound from the "mixsounds"-slot in the
;;; output-stream if "sound" is specified as an integer (1 based
;;; indexing). 
;;;

(defobject sgimixsound (note)
  ((rhythm :initform 0.0)
   sound
   track
   (looping :initform 1)))

(defmethod print-object ((object sgimixsound) stream)
  (printing-random-thing (object stream)
    (formatting-slots (object stream :default +slot-unset+)
                      (time :decimals 3)
                      track
                      looping
		      (sound :printer princ))))

(defmethod write-event ((object sgimixsound) (stream mix-file))
  (let ((file (slot-value stream 'stream)))
    (formatting-slots (object file :preamble "sound " :eol t)
	  (sound :printer princ
		 :filter (lambda (s)
			   (if (integerp s)
			       (nth (max 0 (1- s)) ; begin with 1!
				    (slot-value stream 'mixsounds))
			     s)))
	  (time :decimals 3)
	  track
	  looping)))

;;;
;;; The "variable" object.  The variable settings are actually global,
;;; ie. they affect all 9 tracks simultaneously.  Even so they are
;;; placed around on the local tracks.  This makes a separate
;;; variable-object needed.
;;;

(defobject sgimixvar (note)
  (time
   (rhythm :initform 0.0)
   effect
   track
   value))

(defmethod print-object ((object sgimixvar) stream)
  (printing-random-thing (object stream)
    (formatting-slots (object stream :default +slot-unset+)
        (time :decimals 3)
        track
        (effect :printer princ)
        value)))

(defmethod write-event ((object sgimixvar) (stream mix-file))
  (let ((file (slot-value stream 'stream)))
    (formatting-slots (object file :preamble "variable " :eol t)
        (effect :printer princ)
        track
        (time :decimals 3)
        (value :decimals 3))))

;;; "playsgimix -batch" is a may-be call on mix in batch-mode (5.3.95
;;; not realizable with mix yet. A.V.)

;;;(defmethod play-using-syntax ((syntax mix) file &rest pairs)
;;;  (declare (ignore pairs))
;;;  (let ((command (format nil "~Aunix/playsgimix ~A -batch"
;;;			 *common-music-directory*
;;;			 (namestring (truename file)))))
;;;    (tell-user "Playing: ~A~%" command)
;;;    (cm::shell command)
;;;    file))

#|
;;; in the first example sounds are set inside the algorithm, whilst
;;; in the second theyre looked up from the slot in the mix-file-stream.

;;; "mix" doesnt care about variables and sounds being jammed together
;;; in the script, so its allright to use merges

(merge mix ()
       (algorithm mixsnds1 sgimixsound (end 200 looping 1)
		  (vars (sounds (insounds "andersvi/aka9.aifc, aka10")))
	  (setf rhythm (item (rhythms te te te s s. s. te q w tempo 5)))
	  (setf sound (nth (item (items 0 1 in random)) sounds))
	  (setf track (item (items 2 3 4 5 6 7 8 in heap))))
       (algorithm mixvars1 sgimixvar (end 200)
	  (setf rhythm (item (rhythms w h. tq tq tq tempo 10)))
	  (setf effect (item (items "reverb" "ring" "compressor")))
	  (setf value (item (items 0.0 0.3 0.5)))
	  (setf track 1)))

(merge view ()
   (let (global-rhythm)
     (generator global nil (end 30)
        (setf rhythm (between 0.3 0.8))
	(setf global-rhythm rhythm))
     (generator mixsnds1 sgimixsound (end 30 looping 1)
	(setf rhythm global-rhythm)
	(setf sound (item (items 1 2 3 in heap)))
	(setf track (item (items 2 3 4 in cycle))))
     (generator mixvars1 sgimixvar (end 30)
	(setf rhythm global-rhythm)
	(setf effect (item (items "reverb" "ring" "compressor")))
	(setf value (item (items 0.0 0.3 0.5)))
	(setf track (item (items 5 6 7 in sequence))))))

{open /hf/notam/u1/andersvi/test.mix
        mixsounds (insounds "test, pelle, proffen")}
syn sgimix
write view

|#


